/* Quaternion support */

/* 1995-2002, Martin Strubel // onk@section5.ch
 *
 * $Id: quat.c,v 1.3 2002-08-20 22:20:14 strubi Exp $
 *
 */

#include <math.h>

#include "quat.h"

void quatmulquat(pQuat out, pQuat p, pQuat q)
{
	out[0] = p[0] * q[0] - p[1] * q[1] - p[2] * q[2] - p[3] * q[3];
	out[1] = p[0] * q[1] + p[1] * q[0] + p[2] * q[3] - p[3] * q[2];
	out[2] = p[0] * q[2] - p[1] * q[3] + p[2] * q[0] + p[3] * q[1];
	out[3] = p[0] * q[3] + p[1] * q[2] - p[2] * q[1] + p[3] * q[0];
}

void rotaxis_from_quat(vfloat *angle, pVector3 axis, pQuat q)
{
	vfloat s;
	vfloat phi;

	phi = acos(q[0]);
	if (phi == 0.0) {
		axis[0] = axis[1] = 0.0; axis[2] = 1.0;
		return;
	} else {
		s = 1.0 / sin(phi);
	}
	axis[0] = s * q[1];
	axis[1] = s * q[2];
	axis[2] = s * q[3];
	phi *= 2.0;
	*angle = phi;
}

void quat_from_rotaxis(pQuat out, pVector3 axis, vfloat phi)
{
	vfloat s;
	phi *= 0.5;

	s = sin(phi);
	out[0] = cos(phi);
	out[1] = axis[0] * s;
	out[2] = axis[1] * s;
	out[3] = axis[2] * s;
}

void quat_from_euler(pQuat out, pVector3 eul)
{
	vfloat sx, cx, sy, cy, sz, cz;
	vfloat e;

	e = eul[0] / 2.0; cx = cos(e); sx = sin(e);
	e = eul[1] / 2.0; cy = cos(e); sy = sin(e);
	e = eul[2] / 2.0; cz = cos(e); sz = sin(e);

	out[0] = cx * cy * cz - sx * sy * sz;
	out[1] = sx * cy * cz - cx * sy * sz;
	out[2] = cx * sy * cz + sx * cy * sz;
	out[3] = cx * cy * sz + sx * sy * cz;
}

void euler_from_quat(pVector3 out, pQuat q)
{
	vfloat tmp, r;

	tmp = q[1] * q[1];
	tmp -= q[3] * q[3];
	
	r = (q[0]*q[0] + tmp - q[2] * q[2]);
	out[0] = atan2(2.0 * (q[1] * q[2] + q[0] * q[3]), r);
	out[1] =  asin(2.0 * (q[0] * q[2] - q[1] * q[3]));
	out[2] = atan2(2.0 * (q[0] * q[1] + q[2] * q[3]), (r - 2.0 * tmp));
}

void mat3_from_quat(pMatrix3 m, pQuat q)
{

#define w q[0]
#define x q[1]
#define y q[2]
#define z q[3]

	m[0][0] = 1.0 - 2.0 * (y*y + z*z);
	m[0][1] = 2.0 * (x*y + w*z);
	m[0][2] = 2.0 * (x*z - w*y);
	m[1][0] = 2.0 * (x*y - w*z);
	m[1][1] = 1.0 - 2.0 * (x*x + z*z);
	m[1][2] = 2.0 * (y*z + w*x);
	m[2][0] = 2.0 * (x*z + w*y);
	m[2][1] = 2.0 * (y*z - w*x);
	m[2][2] = 1.0 - 2.0 * (x*x + y*y);
#undef w
#undef x
#undef y
#undef z

}

